/*  This file is a part of Puzzles, a MIDLET for playing puzzles.
    Copyright (C) 2004, Kostas Giannakakis

    Puzzles is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Puzzles is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Puzzles; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
 * @(#)Board.java	1.0 04/02/18
 *					1.1 05/11/16 (Stores/reads board selection from
 *                                permanent storage)
 *
 */

package games.Solo;

import games.*;
import java.util.Date;
import javax.microedition.lcdui.*;
import javax.microedition.midlet.MIDlet;
import javax.microedition.rms.*;
import java.io.IOException;

public class Board extends Canvas implements CommandListener
{

	MIDlet midlet;
	Display dpy;

	/** Table for storing moves. */
	int [] movesTable;

	/** Number of moves so far (0..31).*/
	int moveCount;

	/** Position of selected square 0..48. */
	int selPos;

	/** True if used picked a tile to move. */
	boolean fired;

	/** Table with board layaout. */
	int boardTable[];

	/** Initialisation tables. */
	int initTables[][] ={// Solitaire
						 {-1, -1,  1,  1,  1, -1, -1,
                          -1, -1,  1,  1,  1, -1, -1,
                           1,  1,  1,  1,  1,  1,  1,
                           1,  1,  1,  0,  1,  1,  1,
                           1,  1,  1,  1,  1,  1,  1,
                          -1, -1,  1,  1,  1, -1, -1,
                          -1, -1,  1,  1,  1, -1, -1},
						 // Cross
						 {-1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1,
						   0, 0, 1, 0, 0, 0, 0,
						   0, 1, 1, 1, 1, 0, 0,
						   0, 0, 1, 0, 0, 0, 0,
						  -1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1},
						 // Plus
						 {-1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 0, 1, 0,-1,-1,
						   0, 0, 0, 1, 0, 0, 0,
						   0, 1, 1, 1, 1, 1, 0,
						   0, 0, 0, 1, 0, 0, 0,
						  -1,-1, 0, 1, 0,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1},
						 // Fireplace
						 {-1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1,
						   1, 1, 1, 1, 0, 0, 0,
						   1, 1, 1, 0, 0, 0, 0,
						   1, 1, 1, 1, 0, 0, 0,
						  -1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1},
						 // Pyramid
						 {-1,-1, 0, 0, 1,-1,-1,
						  -1,-1, 0, 1, 1,-1,-1,
						   0, 0, 1, 1, 1, 0, 0,
						   0, 1, 1, 1, 1, 0, 0,
						   0, 0, 1, 1, 1, 0, 0,
						  -1,-1, 0, 1, 1,-1,-1,
						  -1,-1, 0, 0, 1,-1,-1},
						 // Arrow
						 {-1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 1, 0, 0,-1,-1,
						   0, 1, 1, 0, 0, 1, 1,
						   1, 1, 1, 1, 1, 1, 1,
						   0, 1, 1, 0, 0, 1, 1,
						  -1,-1, 1, 0, 0,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1},
						 // Double Arrow
						 {-1,-1, 0, 0, 0,-1,-1,
						  -1,-1, 1, 0, 1,-1,-1,
						   0, 1, 1, 1, 1, 1, 0,
						   1, 1, 1, 1, 1, 1, 1,
						   0, 1, 1, 1, 1, 1, 0,
						  -1,-1, 1, 0, 1,-1,-1,
						  -1,-1, 0, 0, 0,-1,-1},
						 // Diamond
						 {-1,-1, 0, 1, 0,-1,-1,
						  -1,-1, 1, 1, 1,-1,-1,
						   0, 1, 1, 1, 1, 1, 0,
						   1, 1, 1, 0, 1, 1, 1,
						   0, 1, 1, 1, 1, 1, 0,
						  -1,-1, 1, 1, 1,-1,-1,
						  -1,-1, 0, 1, 0,-1,-1}};

	/** Holds number of board selected.*/
	int boardSelection = 0;

	/** Number of boards to select from.*/
	final int BOARD_COUNT = 8;

	/** Peg image. */
	Image pegImage;

	/** Board image. */
	Image boardImage;

	/** Selected peg image. */
	Image pegSelectedImage;

	/** Form for displaying options. */
	SoloOptions optionsForm;

	// commands
	static final int CMD_BOARD_SELECT	= 0;
	static final int CMD_UNDO		= 1;
	static final int CMD_RESET		= 2;
	static final int CMD_BEST_TIME	= 3;
	static final int CMD_HELP		= 4;
	static final int CMD_ABOUT		= 5;
	static final int CMD_RESTART	= 6;
	static final int CMD_EXIT		= 7;
	static final int CMD_ZLAST		= 8;	// must be ze last, of course
	Command cmd[];

	// state variables
	static final int START = 0;
	static final int PLAYING = 1;
	static final int WON = 2;
	static final int ENDED = 3;
	int gameState;

	// Time variables
	Date startTime;
	Date endTime;

	void DebugOut(String s)
	{
		System.out.println(s);
	}

	class BoardCommand extends Command
	{
		int tag;

		BoardCommand(String label, int type, int pri, int tag_)
		{
			super(label, type, pri);
			tag = tag_;
		}
	}

	/**
	 * Initialises board.
	 */
    public void initBoard()
	{
    	resetBoard(true);

    	try
    	{
    		pegImage =
    			Image.createImage("/games/solo/images/peg.png");
    		pegSelectedImage =
    			Image.createImage("/games/solo/images/peg_sel.png");
			boardImage =
	    		Image.createImage("/games/solo/images/board.png");
		}
		catch(IOException e){}
    }

	/**
	 * Resets board's variables.
	 */
    public void resetBoard(boolean resetTime)
	{
    	int i;

    	movesTable = new int[31];
    	for(i =0; i<31; i++)
    		movesTable[i] = -1;
    	moveCount = 0;

		boardTable = new int[49];

    	for(i =0; i<49; i++)
    	{
    		boardTable[i] = initTables[boardSelection][i];
		}

    	selPos = 24;//10;
    	fired = false;

		// set up initial state
		setState(START);
		if (resetTime)
			startTime = new Date();
    }

    public void changeBoard()
    {
		boardSelection = optionsForm.getBoardSelection();
		writeBoardSelection(boardSelection);
		resetBoard(true);
		repaint();
	}

	/**
	 * Sets game's state. Updates commands list.
	 */
	void setState(int ns)
	{
		gameState = ns;
		removeCommand(cmd[CMD_UNDO]);
		removeCommand(cmd[CMD_RESET]);
		removeCommand(cmd[CMD_RESTART]);
		removeCommand(cmd[CMD_BOARD_SELECT]);
		removeCommand(cmd[CMD_BEST_TIME]);
		removeCommand(cmd[CMD_ABOUT]);
		removeCommand(cmd[CMD_HELP]);
		removeCommand(cmd[CMD_EXIT]);
		switch (gameState) {
		case START:
			addCommand(cmd[CMD_BOARD_SELECT]);
			addCommand(cmd[CMD_BEST_TIME]);
			addCommand(cmd[CMD_ABOUT]);
			addCommand(cmd[CMD_HELP]);
			addCommand(cmd[CMD_EXIT]);
			break;
		case PLAYING:
			addCommand(cmd[CMD_UNDO]);
			addCommand(cmd[CMD_RESET]);
			addCommand(cmd[CMD_ABOUT]);
			addCommand(cmd[CMD_HELP]);
			addCommand(cmd[CMD_EXIT]);
			break;
		case WON:
		case ENDED:
			addCommand(cmd[CMD_RESTART]);
			addCommand(cmd[CMD_BOARD_SELECT]);
			addCommand(cmd[CMD_BEST_TIME]);
			addCommand(cmd[CMD_ABOUT]);
			addCommand(cmd[CMD_HELP]);
			addCommand(cmd[CMD_EXIT]);
			break;
		}
	}

	/**
	 * Stores midlet, display variables.
	 * Set up commands, initialises board.
	 */
	public Board(MIDlet midlet_)
	{
		int i;

		// "global" variables
		midlet = midlet_;
		dpy = Display.getDisplay(midlet);

		// set up commands
		cmd = new Command[CMD_ZLAST];

		cmd[CMD_UNDO] =
			new BoardCommand("Undo", Command.SCREEN, 1, CMD_UNDO);

		cmd[CMD_BOARD_SELECT] =
			new BoardCommand("Board", Command.SCREEN, 1, CMD_BOARD_SELECT);

		cmd[CMD_BEST_TIME] =
			new BoardCommand("Best Time", Command.SCREEN, 2, CMD_BEST_TIME);

		cmd[CMD_RESTART] =
			new BoardCommand("Restart", Command.SCREEN, 1, CMD_RESTART);

		cmd[CMD_RESET] =
			new BoardCommand("Reset", Command.SCREEN, 1, CMD_RESET);

		cmd[CMD_ABOUT] =
			new BoardCommand("About", Command.HELP, 3, CMD_ABOUT);

		cmd[CMD_HELP] =
			new BoardCommand("Help", Command.HELP, 3, CMD_HELP);

		cmd[CMD_EXIT] =
			new BoardCommand("Exit", Command.EXIT, 4, CMD_EXIT);

		// set up the listener
		setCommandListener(this);

		// Get latest board selection
		boardSelection = readBoardSelection();

		// Init Board
		initBoard();

		// Initialize Options Form
		optionsForm = new SoloOptions(dpy, this, this);
	}

	/**
	 * Command actions handling.
	 */
	public void commandAction(Command c, Displayable d) {
		switch (((BoardCommand) c).tag) {
			case CMD_ABOUT:
				About.showAbout(Display.getDisplay(midlet));
				break;
			case CMD_HELP:
				HelpSolo.showHelp(Display.getDisplay(midlet));
				break;
			case CMD_EXIT:
				midlet.notifyDestroyed();
				break;
			case CMD_RESTART:
				setState(PLAYING);
				resetBoard(true);
				repaint();
				break;
			case CMD_RESET:
				setState(PLAYING);
				resetBoard(false);
				repaint();
				break;
			case CMD_UNDO:
				undoMove();
				repaint();
				break;
			case CMD_BEST_TIME:
				long bestTime = 0;
				Score score = new Score(BOARD_COUNT);
				if (score.open("SoloScores"))
				{
					bestTime = score.getTime(boardSelection);
				}
				score.close();
				long sec = (bestTime / 1000) % 60,
					 min = (bestTime / 1000) /60;
				String str = new String();
				if (bestTime == 0)
					str = "Not solved";
				else if (sec >= 10)
					str +=  min + ":" + sec;
				else
					str +=  min + ":0" + sec;

				Alert alert = new Alert(optionsForm.getBoardName(boardSelection));
				alert.setTimeout(Alert.FOREVER);
				alert.setString(str);
				dpy.setCurrent(alert);
				break;
			case CMD_BOARD_SELECT:
				optionsForm.setBoardSelection(boardSelection);
				dpy.setCurrent(optionsForm);
				break;
		}
	}

	/**
	 * Paint routine
	 */
    public void paint(Graphics g)
	{
		g.setColor(0xFFFFFF);
		g.fillRect(0, 0, getWidth(), getHeight());

		g.setColor(0);

		int index;
		Image im;

		im = Image.createImage(boardImage);
		g.drawImage(im,
					0,
					0,
					0);


		for(int r = 0; r < 7; r++)
			for(int c = 0; c < 7; c++)
		{
			index = r * 7 + c;

			// Simple board
			/*if (boardTable[index] == 0)
			{
				g.fillArc(4 + c*13,
						  4 + r*13,
						  13,
						  13,
						  0,
						  360);
			}
			else if (boardTable[index] == 1)
			{
				if (fired && index == selPos)
				{
					g.setColor(0x00FF00);
					g.fillArc(4 + c*13,
							  4 + r*13,
							  13,
							  13,
							  0,
							  360);
					g.setColor(0);
				}
				else
				{
					g.drawArc(4 + c*13,
							  4 + r*13,
							  13,
							  13,
							  0,
							  360);
				}
			}*/

			// Graphical board
			if (boardTable[index] == 1)
			{
				if (fired && index == selPos)
					im = Image.createImage(pegSelectedImage);
				else
					im = Image.createImage(pegImage);

				g.drawImage(im,
							4 + c*13,
							4 + r*13,
							0);
			}
		}

		selectSquare(g, selPos / 7, selPos % 7);

		if (gameState == WON)
		{
			Alert alert = new Alert("Well done!");
			alert.setTimeout(Alert.FOREVER);

			long time = endTime.getTime() - startTime.getTime();
			long bestTime = 0;

			// Read best time, update if necessary
			Score score = new Score(BOARD_COUNT);
			if (score.open("SoloScores"))
			{
				bestTime = score.getTime(boardSelection);
				if (bestTime > time || bestTime == 0)
					score.setTime(boardSelection, (int) time);
			}
			score.close();
			optionsForm.updateImageArray();

			long sec = (time / 1000) % 60,
				 min = (time / 1000) /60;

			String str = new String();
			if (bestTime > time || bestTime == 0)
				str = "New best time!\n";
			else
				str = "Solved\nTime: ";
			if (sec >= 10)
				str +=  min + ":" + sec;
			else
				str +=  min + ":0" + sec;
			alert.setString(str);

			Display.getDisplay(midlet).setCurrent(alert);
			gameState = ENDED;

			/*g.translate(-g.getTranslateX(), -g.getTranslateY());
			g.setColor(0xFFFF00);
			g.setFont(Font.getDefaultFont());
			int h = selPos > 35 ? 30 : getHeight() - 1;
			g.drawString("YOU WIN!",
						 getWidth() / 2,
						 h,
						 Graphics.BOTTOM | Graphics.HCENTER);*/
		}
    }

    private void selectSquare(Graphics g,
                              int r,
                              int c)
    {
		int oldColor = g.getColor();

		g.setColor(0xFF0000);
		g.drawRect(3 + c*13,
				   3 + r*13,
				   12,
				   12);
		g.setColor(oldColor);
	}


	/**
	 * Key Pressed handling
	 */
	public void keyPressed(int code)
	{
		if (gameState == WON)
			return;

		int game = getGameAction(code);
		int r = selPos / 7,
		    c = selPos % 7;

		if (!fired)
		{
			switch (game)
			{
				case Canvas.RIGHT:
					c = (c + 1) % 7;
					break;
				case Canvas.DOWN:
					r = (r + 1) % 7;
					break;
				case Canvas.LEFT:
					c = (c > 0) ? (c - 1) : 6;
					break;
				case Canvas.UP:
					r = (r > 0) ? (r - 1) : 6;
					break;
				case Canvas.FIRE:
					if (isMoveLegal(Canvas.RIGHT) ||
						isMoveLegal(Canvas.DOWN) ||
						isMoveLegal(Canvas.LEFT) ||
						isMoveLegal(Canvas.UP))
					{
						fired = true;
						repaint();
						//DebugOut("Selected: " + selPos);
						return;
					}
					break;
				default:
					return;
			}

			int newPos = r*7 + c;

			if (newPos >= 0 && newPos < 49 &&
			    boardTable[newPos] != -1)
			{
				//DebugOut("New pos: " + newPos);
				selPos = newPos;
				repaint();
			}
		}
		else
		{
			switch (game)
			{
				case Canvas.RIGHT:
				case Canvas.DOWN:
				case Canvas.LEFT:
				case Canvas.UP:
					if (isMoveLegal(game))
					{
						makeMove(game);
						fired = false;
						repaint();
						if (isSolved())
						{
							endTime = new Date();
							setState(WON);
						}
					}
					break;
				case Canvas.FIRE:
					fired = false;
					repaint();
					break;
				default:
					return;
			}
		}
	}

	/**
	 * Puzzle's routine. Updates board's table,
	 * registes move, so that undoing is possible.
	 */
    public void makeMove(int dir)
	{
		int removedPosOffsetTbl[] = {1, 7, -1, -7};
		int newPosOffsetTbl[] = {2, 14, -2, -14};
		int dir_ = 0;

		switch (dir)
		{
			case Canvas.RIGHT:
				dir_ = 0;
				break;
			case Canvas.DOWN:
				dir_ = 1;
				break;
			case Canvas.LEFT:
				dir_ = 2;
				break;
			case Canvas.UP:
				dir_ = 3;
				break;
		}

		int pos2 = selPos + removedPosOffsetTbl[dir_],
			pos3 = selPos + newPosOffsetTbl[dir_];

		if (pos2 >=0 && pos2 < 49 &&
			pos3 >=0 && pos3 < 49)
		{
			boardTable[selPos] = 0;
			boardTable[pos2] = 0;
			boardTable[pos3] = 1;

			movesTable[moveCount] = 10*selPos + dir_;
			moveCount++;

			selPos = pos3;

			if (gameState == START)
			{
				setState(PLAYING);
			}
		}
    }

	/**
	 * Puzzle's routine. Undoes last move
	 */
    public void undoMove()
	{
		if (moveCount == 0 ||
		    gameState == WON)
		    return;

		moveCount--;
		int pos = movesTable[moveCount]/10,
		    dir = movesTable[moveCount]%10;
		//int reverseDir = (dir > 1) ? (dir - 2) : (dir + 2);
		//DebugOut("moveCount= " + moveCount + ",Pos= " + selPos + ",dir= " + dir);

		movesTable[moveCount] = -1;

		int removedPosOffsetTbl[] = {1, 7, -1, -7};
		int newPosOffsetTbl[] = {2, 14, -2, -14};
		int pos2 = pos + removedPosOffsetTbl[dir],
			pos3 = pos + newPosOffsetTbl[dir];

		boardTable[pos] = 1;
		boardTable[pos2] = 1;
		boardTable[pos3] = 0;
		selPos = pos;

		if (moveCount == 0)
		{
			setState(START);
		}
    }

	/**
	 * Puzzle's routine. Returns true if move is legal.
	 */
	private boolean isMoveLegal(int dir)
	{
		int r = selPos / 7,
		    c = selPos % 7;

		switch (dir)
		{
			case Canvas.RIGHT:
				if (c >= 5 ||
					boardTable[selPos+1] != 1 ||
					boardTable[selPos+2] != 0)
					return(false);
				break;
			case Canvas.DOWN:
				if (r >= 5 ||
					boardTable[selPos+7] != 1 ||
					boardTable[selPos+14] != 0)
					return(false);
				break;
			case Canvas.LEFT:
				if (c <= 1 ||
					boardTable[selPos-1] != 1 ||
					boardTable[selPos-2] != 0)
					return(false);
				break;
			case Canvas.UP:
				if (r <= 1 ||
					boardTable[selPos-7] != 1 ||
					boardTable[selPos-14] != 0)
					return(false);
			break;
			default:
				return(false);
		}
		return(true);
	}


	/**
	 * Puzzle's routine. Returns true if the puzzle is solved.
	 */
	private boolean isSolved()
	{
    	int i, count = 0;
    	for(i =0; i<49; i++)
    	{
    		if (boardTable[i] == 1)
    		{
    			count++;
    			if (count > 1)
    				return(false) ;
			}
		}
		return(true);
	}

	/**
	 * Reads board selection stored in permanent storage.
	 */
	private int readBoardSelection()
	{
		RecordStore store = null;
		int boardSelection = 0;

		try {
			store = RecordStore.openRecordStore("BoardSelection", true);
		} catch (RecordStoreException ex) {
		}

		if (store == null)
			return 0;

		try
		{
			RecordEnumeration enum = store.enumerateRecords(null, null, false);
			if (enum.hasNextElement())
			{
				int index = enum.nextRecordId();
				byte buf[] = new byte[1];
				int length = store.getRecord(index, buf, 0);
				if (length == 1)
				{
					boardSelection = (int)  buf[0];
				}
				store.closeRecordStore();
			}
		} catch (RecordStoreException ex) {
			ex.printStackTrace();
		}

		return boardSelection;
	}

	/**
	 * Writes board selection in permanent storage.
	 */
	private void writeBoardSelection(int boardSelection)
	{
		RecordStore store = null;

		try {
			store = RecordStore.openRecordStore("BoardSelection", true);
		} catch (RecordStoreException ex) {
		}

		if (store == null)
			return;

		byte buf[] = new byte[1];
		buf[0] = (byte) boardSelection;

		try
		{
			RecordEnumeration enum = store.enumerateRecords(null, null, false);
			if (enum.hasNextElement())
			{
				int index = enum.nextRecordId();
				store.setRecord(index, buf, 0, 1);
			}
			else
			{
				store.addRecord(buf, 0, 1);
			}
			store.closeRecordStore();
		}
		catch (RecordStoreException ex)
		{
		    System.out.println("RecordStoreException");
		    ex.printStackTrace();
		}
	}
}
