/*  This file is a part of Puzzles, a MIDLET for playing puzzles.
    Copyright (C) 2004, Kostas Giannakakis

    Puzzles is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Puzzles is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Puzzles; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
 * @(#)SoloOptions.java   1.0 04/07/27
 *
 */

package games.Solo;

import games.Score;
import javax.microedition.lcdui.*;

public class SoloOptions extends Form implements CommandListener
{
	ChoiceGroup group;

	Command ok;
	Command cancel;

	Display dpy;
	Displayable prev;
	Board board;

	static final int BOARD_COUNT = 8;
	private int boardSelection = 0;
	private Image im1, im2;
	private Image[] imageArray;
	private	String[] stringArray = {
			"Solitaire", "Cross", "Plus", "Fireplace",
			"Pyramid", "Arrow", "Double Arrow", "Diamond"};
	private boolean first;

    public SoloOptions(Display dpy, Displayable prev, Board board)
    {
        super("Board");

        this.dpy = dpy;
        this.prev = prev;
        this.board = board;

		try
		{
			im1 = Image.createImage("/games/solo/images/peg_sel.png");
			im2 = Image.createImage("/games/solo/images/peg.png");
		}
		catch(java.io.IOException err) {}

		imageArray = new Image[BOARD_COUNT];
		first = true;

		updateImageArray();

		ok = new Command("Back", Command.BACK, 0);
		//cancel = new Command("Cancel", Command.CANCEL, 1);

		addCommand(ok);
		//addCommand(cancel);
		setCommandListener(this);
    }

	public void updateImageArray()
	{
		Score score = new Score(BOARD_COUNT);
		if (score.open("SoloScores"))
		{
			for(int i=0;i<BOARD_COUNT;i++)
			{
				if (score.getStatus(i))
					imageArray[i] = im1;
				else
					imageArray[i] = im2;
			}
		}
		score.close();

		// create the list of choice groups.
		if (first)
		{
			group = new ChoiceGroup("Available boards",
									ChoiceGroup.EXCLUSIVE,
									stringArray,
									imageArray);
			append(group);
			first = false;
		}
		else
		{
			try
			{
				for(int i=0;i<BOARD_COUNT;i++)
				{
					group.set(i, stringArray[i], imageArray[i]);
				}
			}
			catch (IndexOutOfBoundsException ex1) {}
			catch (NullPointerException ex2){}
		}
	}

	public void commandAction(Command c, Displayable d)
	{
		if (c == ok)
		{
			boardSelection = group.getSelectedIndex();
			board.changeBoard();
		}
		/*else if (c == cancel)
		{
			setBoardSelection(boardSelection);
		}*/
		dpy.setCurrent(prev);
	}

	public int getBoardSelection()
	{
		//boardSelection = group.getSelectedIndex();
		return(boardSelection);
	}

	public void setBoardSelection(int sel)
	{
		if (sel < BOARD_COUNT)
		{
			boardSelection = sel;
			for(int i=0; i<BOARD_COUNT; i++)
				group.setSelectedIndex(i, i == sel);
		}
	}

	public String getBoardName(int sel)
	{
		if (sel < BOARD_COUNT)
		{
			return stringArray[sel];
		}
		return "Error";
	}
}